unit FARUtils;

interface
{$ALIGN OFF}
{$MINENUMSIZE 4}

uses
  FARIntf, Dirs, SysUtils, Classes;

function iif(Condition: boolean; ReturnIfTrue, ReturnIfFalse: variant):variant;
function FindFirstNonSpace(s:string):integer;
function PluginDir:string;
function PluginDirFile(const s:string):string;
function EditorFileFullPath:string;
function EditorFileExt:string;
function EditorFileID:integer;

function newFARMenuItems(Size:integer):PFARMenuItemsArray;overload;
function showFARMenu(Title:PChar; Items:PFARMenuItemsArray; ItemsSize:Integer):integer;overload;
function newFARMenuItems(s:TStrings):PFARMenuItemsArray;overload;
function newFARMenuItemsFromNames(s:TStrings):PFARMenuItemsArray;overload;
function showFarMenu(Title:PChar;S:TStrings):integer;overload;
procedure freeFARMenuItems(a:PFARMenuItemsArray; Size:integer);
procedure FARExec(const s:string);
procedure farSaveFile;
function  ExpandEnv(const s:string):string;
function SelectString(Title:PChar;S:TStrings):String;
procedure SetEditorPosition(CurLine, CurPos, CurTabPos, TopScreenLine, LeftPos: Integer;  Overtype: LongBool);
procedure Edit(const FileName, Title:string;StartLine, StartChar:integer);
function hasSelection:boolean;
function Selection:String;
function currentWord:String;
function getWord:String;
implementation
uses ShellApi,  Windows;
function iif(Condition:boolean; ReturnIfTrue,ReturnIfFalse:variant):variant;
  begin
    if Condition then result:=ReturnIfTrue else result:=ReturnIfFalse;
  end;
function FindFirstNonSpace(s:string):integer;
  begin
    Result:=1;
    while (Result<Length(s)) and (s[Result] in [#32,#9]) do begin
      Inc(Result);
    end;
  end;
function PluginDir:string;
  begin
    result:=getFileDir(info.moduleName);
  end;
function PluginDirFile(const s:string):string;        
  begin
    result:=DirFile(PluginDir, s);
  end;
function EditorFileFullPath:string;
  Var einfo:TEditorInfo;
  begin
    info.EditorControl(ECTL_GETINFO, @einfo);
    Result:=getFullPath(einfo.FileName);
  end;
function EditorFileID:integer;
Var einfo:TEditorInfo;
begin
  info.EditorControl(ECTL_GETINFO, @einfo);
  Result:=einfo.EditorID;
end;

function EditorFileExt:string;
  Var einfo:TEditorInfo;
  begin
    info.EditorControl(ECTL_GETINFO, @einfo);
    Result:=ExtractFileExt(einfo.FileName);
  end;
function newFARMenuItems(Size:integer):PFARMenuItemsArray;
         begin
           GetMem(Result, Size*SizeOf(TFarMenuItem));
           FillChar(Result^, Size*SizeOf(TFarMenuItem), 0);
         end;
function showFARMenu(Title:PChar; Items:PFARMenuItemsArray; ItemsSIze:Integer):integer;overload;
        begin
          result:=Info.Menu(Info.ModuleNumber,-1,-1,0,
                         FMENU_AUTOHIGHLIGHT,
                         Title,
                         nil,
                         'Contents',
                         nil,nil,
                         Items,ItemsSize);

        end;
function newFARMenuItems(s:TStrings):PFARMenuItemsArray;overload;
         Var i:integer;
         begin
           result:=newFARMenuItems(s.Count);
           for i:=0 to s.Count-1 do
           begin
             StrLCopy(Result[i].Text, PChar(s[i]), SizeOf(Result[i].Text)-1);
           end;
         end;
procedure freeFARMenuItems(a:PFARMenuItemsArray; Size:integer);
         begin
           freeMem(a, Size);
         end;
function newFARMenuItemsFromNames(s:TStrings):PFARMenuItemsArray;overload;
         Var i:integer;
         begin
           result:=newFARMenuItems(s.Count);
           for i:=0 to s.Count-1 do
           begin
             if S[i]='---' then
               Result[i].Separator:=true;
             StrLCopy(Result[i].Text, PChar(s.Names[i]), SizeOf(Result[i].Text)-1);
           end;
         end;
function hasSelection:boolean;
Var gs:TEditorGetString;
begin
  gs.StringNumber:=-1;
  info.EditorControl(ECTL_GETString,@gs);
  Result:=gs.SelStart<>-1;
end;
function Selection:String;
Var gs:TEditorGetString;
    c:PChar;
    pos:integer;
begin
  gs.StringNumber:=-1;
  info.EditorControl(ECTL_GETString,@gs);
  if gs.SelEnd=-1 then
  begin
    Result:='';
  end
  else
  begin
    c:=gs.StringText;
    inc(c,gs.SelStart);
    pos:=gs.SelStart;
    while pos<gs.SelEnd do
    begin
      Result:=Result+c^;
      inc(c);
      inc(pos);
    end;
  end;
end;
const Spaces=[' ',#8];
function currentWord:String;
Var gs:TEditorGetString;
    ei:TEditorInfo;
    c:PChar;
    pos:integer;
begin
  gs.StringNumber:=-1;
  info.EditorControl(ECTL_GETString,@gs);
  info.EditorControl(ECTL_GETInfo,@ei);
  c:=gs.StringText;
  pos:=ei.CurPos;
  if pos>=gs.StringLength then
    pos:=gs.StringLength-1;
  inc(c,pos);
  while (not (c^ in Spaces)) and (c<>gs.StringText) do
  begin
    dec(c);
    dec(pos);
  end;
  while (c^ in Spaces) and (pos<(gs.StringLength-1)) do
  begin
    inc(c);
    inc(pos);
  end;
  Result:='';
  while (not (c^ in Spaces)) and (pos<(gs.StringLength)) do
  begin
    Result:=Result+c^;
    inc(c);
    inc(pos);
  end;
end;
function getWord:String;
begin
  if hasSelection then
    result:=selection
  else
    result:=currentWord;
end;

procedure BlankScreen;
Var Blank:array[0..1024] of char;
    csbi:CONSOLE_SCREEN_BUFFER_INFO;
    Y:Integer;
    C:COORD;
    StdOutput:integer;
begin
    StdOutput:=GetStdHandle(STD_OUTPUT_HANDLE);
    GetConsoleScreenBufferInfo(StdOutput,csbi);
    FillChar(Blank, csbi.dwSize.X, ' ');
    Blank[csbi.dwSize.X]:=#0;
    for Y:=0 to csbi.dwSize.Y-1 do
      Info.Text(0,Y,8,Blank);
    Info.Text(0,0,0,nil);

    C.X:=0;
    C.Y:=csbi.dwCursorPosition.Y;
    SetConsoleCursorPosition(StdOutput,C);
end;
const MaxCommandLength=255;
function  ExpandEnv;
Var expanded:array [0..MaxCommandLength] of char;
begin
  ExpandEnvironmentStrings(PChar(s),Expanded,MaxCommandLength);
  Result:=expanded;
end;
procedure FARExec(const s:string);
         Var h:integer;
             si:STARTUPINFO;
             pi:PROCESS_INFORMATION;
             scr:integer;
             msg:array [0..200] of char;
             oldtitle:array [0..512] of char;
         begin
           scr:=info.SaveScreen(0,0,-1,-1);
           FillChar(si,sizeof(si), 0);
           si.cb:=sizeof(si);
           BlankScreen;
           GetConsoleTitle(oldTitle,sizeof(oldTitle));
           SetConsoleTitle(PChar(s));
           writeln('>"', PChar(s), '"');
           if CreateProcess(nil, PChar(s),nil,nil,false,0,nil,nil,
                si,
                pi
              )
           then
           begin
             h:=pi.hProcess;
             WaitForSingleObject(h,INFINITE);
           end
           else
           begin
             FormatMessage(
                  FORMAT_MESSAGE_FROM_SYSTEM,      // source and processing options 
                nil,   // pointer to  message source 
                GetLastError,  // requested message identifier 
                0, // language identifier for requested message 
                msg,    // pointer to message buffer 
                200,        // maximum size of message buffer 
                nil  // address of array of message inserts 
             );
             charToOem(msg, msg);
             writeln(msg);
           end;
           SetConsoleTitle(oldTitle);
           Info.Control(Info.ModuleNumber, FCTL_SETUSERSCREEN, nil);
           info.RestoreScreen(0);
           info.RestoreScreen(scr);
         end;
procedure farSaveFile;
          Var a:TEditorSaveFile;
          begin
            a.FileName[0]:=#0;
            a.FileEol:=nil;
            Info.EditorControl(ECTL_SaveFile, @a)
          end;
function showFarMenu(Title:PChar;S:TStrings):integer;overload;
         Var menuItems:PFARMenuItemsArray;
         begin
           menuItems:=newFARMenuItems(s);
           try
             result:=showFARMenu(Title, menuItems, s.Count);
           finally
             freeFARMenuItems(menuItems, s.Count);
           end;
         end;
function SelectString(Title:PChar;S:TStrings):String;
Var item:integer;
begin
 item:=showFarMenu(Title, S);
 if item>=0 then
   Result:=S[item]
 else
   result:='';
end;
procedure SetEditorPosition(CurLine, CurPos, CurTabPos, TopScreenLine, LeftPos: Integer; Overtype: LongBool);
Var p:TEditorSetPosition;
begin
  p.CurLine:=CurLine;
  p.CurPos:=CurPos;
  p.CurTabPos:=CurTabPos;
  p.TopScreenLine:=TopScreenLine;
  p.LeftPos:=LeftPos;
  p.Overtype:=Overtype;
  info.EditorControl(ECTL_SETPOSITION, @p);
end;

procedure Edit(const FileName,Title:string; StartLine,StartChar:integer);
var PTitle: PChar;
begin //here StartLine and StartChar come beginning from 1
  if (Title='') then PTitle:=Nil else PTitle:=PChar(Title);
  info.Editor(PChar(FileName),PTitle, 0,0,-1,-1, 1{EF_NONMODAL}, StartLine,iif(StartChar<=0,1,StartChar));
end;
end.
